{-# LANGUAGE TupleSections, LambdaCase #-}

module Parsing (Parser, eof, ws, char, string, noneOf, digit, chainl1, (<|>), many, some, optional, runParser) where

import Control.Applicative
import Control.Monad
import Data.Char
import Data.Maybe

newtype Parser a = Parser { runParser :: String -> Maybe (String, a) }

instance Functor Parser where
  fmap f (Parser g) = Parser (fmap (fmap f) . g)

instance Applicative Parser where
  pure x = Parser (Just . (,x))
  (Parser f) <*> (Parser x) = Parser $ \s -> do
    (s', g) <- f s
    (s'', y) <- x s'
    return (s'', g y)

instance Alternative Parser where
  empty = Parser (const Nothing)
  (Parser x) <|> (Parser y) = Parser $ (<|>) <$> x <*> y

instance Monad Parser where
  (Parser m) >>= f = Parser $ (>>= uncurry (flip (runParser . f))) . m

satisfy :: (Char -> Bool) -> Parser Char
satisfy f = Parser $ \case
  (x:xs) | f x -> Just (xs, x)
  _ -> Nothing

eof :: Parser ()
eof = Parser $ \s -> if null s then Just (s, ()) else Nothing

ws :: Parser ()
ws = void $ many (satisfy isSpace)

char :: Char -> Parser Char
char = satisfy . (==)

string :: String -> Parser String
string = foldr (liftA2 (:)) (pure []) . map char

noneOf :: [Char] -> Parser Char
noneOf = satisfy . (not .) . flip elem

digit :: Parser Char
digit = satisfy isDigit

chainl1 :: Parser a -> Parser (a -> a -> a) -> Parser a
chainl1 p op = p >>= go
  where
    go x =
      (do f <- op
          y <- p
          go (f x y)
      ) <|> return x
